import unittest
import pyspedas
from pytplot import data_exists
from pyspedas.mms.cotrans.mms_qcotrans import mms_qcotrans

class CotransTestCases(unittest.TestCase):
    def test_qcotrans_sm_to_gse(self):
        pyspedas.mms.mec()
        mms_qcotrans('mms1_mec_v_sm', 'mms1_mec_v_sm_2gse', out_coord='gse')
        mms_qcotrans('mms1_mec_r_sm', 'mms1_mec_r_sm_2gse', out_coord='gse')
        self.assertTrue(data_exists('mms1_mec_v_sm_2gse'))
        self.assertTrue(data_exists('mms1_mec_r_sm_2gse'))

if __name__ == '__main__':
    unittest.main()



from .load import load

def mag(trange=['2017-03-27/06:00', '2017-03-27/08:00'],
        probe='a',
        datatype='hr',
        level='l1b', 
        suffix='', 
        varnames=[],
        time_clip=False):
    """
    This function loads data from the Vector Field Magnetometer (VFM)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Swarm spacecraft ID ('a', 'b' and/or 'c')

        datatype: str
            Data type; Valid options:
                'hr', 'lr'

        level: str
            Data level; options: 'l1b'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='mag', trange=trange, probe=probe, level=level, datatype=datatype, suffix=suffix, varnames=varnames, time_clip=time_clip)



import os

CONFIG = {'local_data_dir': 'stereo_data/',
          'remote_data_dir': 'http://sprg.ssl.berkeley.edu/data/misc/stereo/'}

# override local data directory with environment variables
if os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir'] = os.sep.join([os.environ['SPEDAS_DATA_DIR'], 'stereo'])

if os.environ.get('STEREO_DATA_DIR'):
    CONFIG['local_data_dir'] = os.environ['STEREO_DATA_DIR']


import logging
from viresclient import SwarmRequest
from pyspedas import time_datetime
from pytplot import store_data, options


def load(trange=None,
         collection=None,
         measurements=None,
         models=None,
         sampling_step=None,
         auxiliaries=None,
         residuals=False):
    """

    """
    if trange is None:
        logging.error('No time range specified')
        return

    tr = time_datetime(trange)

    if not isinstance(measurements, list):
        measurements = [measurements]

    if auxiliaries is not None and not isinstance(auxiliaries, list):
        auxiliaries = [auxiliaries]

    if models is not None:
        if not isinstance(models, list):
            models = [models]

    request = SwarmRequest()
    if isinstance(collection, list):
        request.set_collection(*collection)
    else:
        request.set_collection(collection)

    if auxiliaries is None:
        request.set_products(measurements=measurements, models=models, sampling_step=sampling_step, residuals=residuals)
    else:
        request.set_products(measurements=measurements, auxiliaries=auxiliaries, models=models, sampling_step=sampling_step, residuals=residuals)

    data = request.get_between(start_time=tr[0], end_time=tr[1])
    return xarray_to_tplot(data.as_xarray())


def xarray_to_tplot(xr):
    out = []
    for key in xr.keys():
        times = xr[key].coords['Timestamp'].to_numpy()
        saved = store_data(key, data={'x': times, 'y': xr[key].data})
        options(key, 'ytitle', xr[key].description)
        options(key, 'ysubtitle', '[' + xr[key].units + ']')

        # find the legend if this is a vector
        for item in xr[key].coords:
            if item != 'Timestamp':
                options(key, 'legend_names', xr[key].coords[item].values.tolist())

        if saved:
            out.append(key)
        else:
            logging.warning('Problem saving: ' + key)
    return out



from .load import load, loadr
from .load_orbit import load_orbit


def orbit(trange=['2013-11-5', '2013-11-6'],
          probe='15',
          prefix='',
          suffix='',
          downloadonly=False,
          no_update=False,
          notplot=False,
          get_support_data=False,
          varformat=None,
          varnames=[],
          time_clip=True):
    """

    This function loads GOES orbit data (probes 8-18)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            If set, load the data into dictionaries containing the numpy objects instead
            of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load_orbit(trange=trange, probe=probe, varnames=varnames, varformat=varformat,
                      prefix=prefix, suffix=suffix, downloadonly=downloadonly,
                      no_update=no_update, time_clip=time_clip, notplot=notplot, get_support_data=get_support_data)


def fgm(trange=['2013-11-5', '2013-11-6'],
        probe='15',
        datatype='1min',
        prefix='',
        suffix='',
        downloadonly=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the GOES Magnetometer  (probes 8-15)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        datatype: str
            Data type; Valid options:

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='fgm', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix, downloadonly=downloadonly,
                time_clip=time_clip, no_update=no_update)


def eps(trange=['2013-11-5', '2013-11-6'],
        probe='12',
        datatype='1min',
        prefix='',
        suffix='',
        downloadonly=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the GOES energetic particle sensor  (probes 8-15)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        datatype: str
            Data type; Valid options:

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='eps', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def epead(trange=['2013-11-5', '2013-11-6'],
          probe='15',
          datatype='1min',
          prefix='',
          suffix='',
          downloadonly=False,
          no_update=False,
          time_clip=False):
    """
    This function loads data from the GOES Electron, Proton, Alpha Detector  (probes 8-15)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        datatype: str
            Data type; Valid options:

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='epead', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def maged(trange=['2013-11-5', '2013-11-6'],
          probe='15',
          datatype='1min',
          prefix='',
          suffix='',
          downloadonly=False,
          no_update=False,
          time_clip=False):
    """
    This function loads data from the GOES Magnetospheric Electron Detector (probes 8-15)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        datatype: str
            Data type; Default '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='maged', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def magpd(trange=['2013-11-5', '2013-11-6'],
          probe='15',
          datatype='1min',
          prefix='',
          suffix='',
          downloadonly=False,
          no_update=False,
          time_clip=False):
    """
    This function loads data from the GOES Magnetospheric Proton Detector (probes 8-15)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        datatype: str
            Data type; Default '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='magpd', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def hepad(trange=['2013-11-5', '2013-11-6'],
          probe='15',
          datatype='1min',
          prefix='',
          suffix='',
          downloadonly=False,
          no_update=False,
          time_clip=False):
    """
    This function loads data from the GOES High energy Proton and Alpha Detector (probes 8-15)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        datatype: str
            Data type; Default '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='hepad', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def xrs(trange=['2013-11-5', '2013-11-6'],
        probe='15',
        datatype='1min',
        prefix='',
        suffix='',
        downloadonly=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the GOES X-ray Sensor (probes 8-18)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        datatype: str
            Data type; Default '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='xrs', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def euvs(trange=['2023-01-30', '2023-01-31'],
         probe='16',
         datatype='1min',
         prefix='',
         suffix='',
         downloadonly=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the GOES Extreme Ultraviolet Sensor (EUVS), (probes 16-18)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=16

        datatype: str
            Data type; Default '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='euvs', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def mag(trange=['2023-01-30', '2023-01-31'],
        probe='16',
        datatype='1min',
        prefix='',
        suffix='',
        downloadonly=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the GOES Magnetometer, (probes 16-18)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=16

        datatype: str
            Data type; Default '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='mag', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def mpsh(trange=['2023-01-30', '2023-01-31'],
         probe='16',
         datatype='1min',
         prefix='',
         suffix='',
         downloadonly=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the GOES Magnetospheric Particle Sensor (MPS-HI), (probes 16-18)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=16

        datatype: str
            Data type; Default '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='mpsh', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)


def sgps(trange=['2023-01-30', '2023-01-31'],
         probe='16',
         datatype='1min',
         prefix='',
         suffix='',
         downloadonly=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the GOES Solar and Galactic Proton Sensor (SGPS), (probes 16-18)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=16

        datatype: str
            Data type; Default '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    return load(instrument='sgps', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix,
                downloadonly=downloadonly, time_clip=time_clip, no_update=no_update)



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG

def load(trange=['2018-11-5', '2018-11-6'], 
         probe='1',
         instrument='lad', 
         datatype='', 
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads TWINS data; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.twins.lad
        pyspedas.twins.ephemeris
        pyspedas.twins.imager

    """

    if not isinstance(probe, list):
        probe = [probe]

    probe = [str(prb) for prb in probe]

    out_files = []

    for prb in probe:
        if instrument == 'lad':
            pathformat = 'twins'+prb+'/'+instrument+'/%Y/twins'+prb+'_l1_lad_%Y%m%d_v??.cdf'
        elif instrument == 'imager':
            pathformat = 'twins'+prb+'/'+instrument+'/%Y/twins'+prb+'_l1_imager_%Y%m%d??_v??.cdf'
        elif instrument == 'ephemeris':
            pathformat = 'twins'+prb+'/'+instrument+'/'+datatype+'/%Y/twins'+prb+'_'+datatype+'_def_%Y%m%d_v??.cdf'

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars


"""
Transform datetimes from string to decimal.

Examples
--------
time_float()
time_float(['2017-06-29 23:59:59.1234', '2017-12-01 00:15:21.1234'])
time_double()

Notes
-----
Similar to time_double.pro in IDL SPEDAS.

"""
import pytplot
import logging

def time_float_one(s_time=None):
    """
    Transform one datetime from string to decimal.

    Parameters
    ----------
    s_time : str, optional
        Input string.
        The default is None, which returns Now.

    Returns
    -------
    float
        Output time.

    """
    logging.info("time_float_one has been moved to the pytplot package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.time_float_one(s_time=s_time)

def time_float(str_time=None):
    """
    Transform a list of datetimes from string to decimal.

    Parameters
    ----------
    str_time : str/list of str, optional
        Input times. The default is None.

    Returns
    -------
    list of float
        Output times as floats.

    """
    logging.info("time_float has been moved to the pytplot package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.time_float(str_time=str_time)

def time_double(str_time=None):
    """
    Transform a list of datetimes from string to decimal.

    Same as time_float.

    Parameters
    ----------
    str_time : str/list of str, optional
        Input times. The default is None.

    Returns
    -------
    list of float
        Output times as floats.

    """
    logging.info("time_double has been moved to the pytplot package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.time_float(str_time=str_time)


import unittest
from ..feeps.mms_read_feeps_sector_masks_csv import mms_read_feeps_sector_masks_csv
from pyspedas import mms_load_feeps, mms_feeps_pad
from pyspedas.mms.feeps.mms_feeps_gpd import mms_feeps_gpd
from pytplot import del_data, tplot, data_exists


class FEEPSTestCases(unittest.TestCase):
    def test_feeps_pad_regression(self):
        """
        This is a regression test for a bug caused by the v7 of the FEEPS CDF files
        v7 CDFs have an extra dimension on the variables containing the pitch angle data
        """
        mms_load_feeps(datatype='electron', trange=['2016-11-23', '2016-11-24'], data_rate='srvy', probe=4)
        mms_feeps_pad(probe=4)
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad'))
        tplot('mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad', display=False)
        del_data('*')

    def test_gyrophase_angles(self):
        mms_load_feeps(data_rate='brst', trange=['2017-07-11/22:34', '2017-07-11/22:34:25'], probe=3)
        mms_feeps_gpd(probe='3', energy=[61, 77], data_rate='brst')
        self.assertTrue(data_exists('mms3_epd_feeps_brst_l2_electron_intensity_61-77keV_gpd'))
        mms_feeps_gpd(probe='3', data_rate='brst')
        self.assertTrue(data_exists('mms3_epd_feeps_brst_l2_electron_intensity_50-500keV_gpd'))
        tplot(['mms3_epd_feeps_brst_l2_electron_intensity_61-77keV_gpd',
               'mms3_epd_feeps_brst_l2_electron_intensity_50-500keV_gpd'], display=False)

    def test_pad_ions_brst(self):
        mms_load_feeps(probe=4, data_rate='brst', datatype='ion', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        mms_feeps_pad(probe=4, data_rate='brst', datatype='ion', angles_from_bfield=True)
        self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_ion_intensity_70-600keV_pad'))
        self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_ion_intensity_70-600keV_pad_spin'))
        tplot(['mms4_epd_feeps_brst_l2_ion_intensity_70-600keV_pad',
               'mms4_epd_feeps_brst_l2_ion_intensity_70-600keV_pad_spin'], display=False)

    def test_pad_ions_srvy(self):
        mms_load_feeps(probe=4, datatype='ion', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        mms_feeps_pad(probe=4, datatype='ion')
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_ion_intensity_70-600keV_pad'))
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_ion_intensity_70-600keV_pad_spin'))
        tplot(['mms4_epd_feeps_srvy_l2_ion_intensity_70-600keV_pad',
               'mms4_epd_feeps_srvy_l2_ion_intensity_70-600keV_pad_spin'], display=False)

    def test_pad_electrons_srvy(self):
        mms_load_feeps()
        mms_feeps_pad()
        self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad_spin'))
        self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad'))
        tplot(['mms1_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad',
               'mms1_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad_spin'], display=False)

    def test_pad_electrons_srvy_probe(self):
        mms_load_feeps(probe=4)
        mms_feeps_pad(probe=4)
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad_spin'))
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad'))
        tplot(['mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad',
               'mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad_spin'], display=False)

    def test_sector_masks(self):
        d = mms_read_feeps_sector_masks_csv(['2015-08-01', '2015-08-02'])
        self.assertTrue(d['mms4imaskt2'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 58, 59, 60, 61, 62, 63])
        self.assertTrue(d['mms3imaskb3'] == [18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48])
        self.assertTrue(d['mms4imaskt8'] == [0, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 57, 58, 59, 61])
        self.assertTrue(d['mms4imaskb4'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63])
        self.assertTrue(d['mms4imaskt1'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63])
        self.assertTrue(d['mms1imaskt2'] == [11, 12])
        self.assertTrue(d['mms3imaskb1'] == [20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43])
        self.assertTrue(d['mms3imaskb5'] == [36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61])
        self.assertTrue(d['mms1imaskt3'] == [22, 23, 24, 25, 26, 27, 28, 31, 32])
        d = mms_read_feeps_sector_masks_csv(['2016-08-01', '2016-08-02'])
        self.assertTrue(d['mms1imaskb4'] == [20, 21, 22, 23, 24, 25, 26, 27, 28, 37, 38, 47])
        self.assertTrue(d['mms4imaskb9'] == [31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 53, 54, 55, 56])
        self.assertTrue(d['mms3imaskb9'] == [53, 54, 55, 56, 57])
        self.assertTrue(d['mms4imaskb7'] == [30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43])
        self.assertTrue(d['mms3imaskb1'] == [20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43])
        self.assertTrue(d['mms4imaskt9'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 25, 26, 27, 28, 29, 30, 31, 32, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63])
        self.assertTrue(d['mms3imaskb8'] == [31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45])
        d = mms_read_feeps_sector_masks_csv(['2017-08-01', '2017-08-02'])
        self.assertTrue(d['mms3imaskb9'] == [53, 54, 55, 56, 57, 58])
        self.assertTrue(d['mms2imaskt9'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 53, 54, 60, 61, 62, 63])
        self.assertTrue(d['mms1imaskb6'] == [40, 41, 42, 49, 50, 51, 52, 53, 54, 57, 58])


if __name__ == '__main__':
    unittest.main()



"""
Find all pytplot names that follow a pattern that may contain a wildcard.

Notes
-----
Allowed wildcards are ? for a single character, * from multiple characters.

"""
import logging
import pytplot

def tnames(pattern=None):
    """
    Find pytplot names.

    Parameters
    ----------
    pattern : str, optional
        Patern to search for. The default is None, which returns all names.

    Returns
    -------
    name_list : list of str
        List of pytplot variables.

    """
    logging.info("tnames has been moved to the pytplot package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.tnames(pattern)



import os
import unittest
from pytplot import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_utc_timestamp_regression(self):
        varname = 'BX_GSE'
        data_omni = pyspedas.omni.data(trange=['2010-01-01/00:00:00', '2010-01-02/00:00:00'],notplot=True,varformat=varname,time_clip=True)
        self.assertTrue(str(data_omni[varname]['x'][0]) == '2010-01-01 00:00:00')

    def test_load_hro2_data(self):
        omni_vars = pyspedas.omni.data()
        self.assertTrue(data_exists('BX_GSE'))
        self.assertTrue(data_exists('BY_GSE'))
        self.assertTrue(data_exists('BZ_GSE'))
        self.assertTrue(data_exists('BY_GSM'))
        self.assertTrue(data_exists('BZ_GSM'))
        self.assertTrue(data_exists('proton_density'))

    def test_load_hro_data(self):
        omni_vars = pyspedas.omni.data(level='hro')
        self.assertTrue(data_exists('BX_GSE'))
        self.assertTrue(data_exists('BY_GSE'))
        self.assertTrue(data_exists('BZ_GSE'))
        self.assertTrue(data_exists('BY_GSM'))
        self.assertTrue(data_exists('BZ_GSM'))
        self.assertTrue(data_exists('proton_density'))

    def test_load_hro_5min_data(self):
        omni_vars = pyspedas.omni.data(level='hro', datatype='5min')
        self.assertTrue(data_exists('BX_GSE'))
        self.assertTrue(data_exists('BY_GSE'))
        self.assertTrue(data_exists('BZ_GSE'))
        self.assertTrue(data_exists('BY_GSM'))
        self.assertTrue(data_exists('BZ_GSM'))
        self.assertTrue(data_exists('proton_density'))

    def test_load_hro_hour_data(self):
        omni_vars = pyspedas.omni.data(level='hro2', datatype='hour', trange=['2013-01-01', '2013-01-02'])

    def test_load_invalid_datatype(self):
        omni_vars = pyspedas.omni.data(datatype='1')

    def test_downloadonly(self):
        files = pyspedas.omni.data(downloadonly=True, trange=['2014-2-15', '2014-2-16'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



from .load import load
from pyspedas.utilities.datasets import find_datasets


def mag(trange=['1983-02-16', '1983-02-17'],
        datatype='62ms',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Magnetometer (MAG)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options: '62ms'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """

    tvars = load(instrument='mag', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return mag_postprocessing(tvars)


def mag_postprocessing(variables):
    """
    Placeholder for MAG post-processing 
    """
    return variables


def nacs(trange=['1983-02-01', '1983-02-02'],
        datatype='neutral1s',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Neutral Atmosphere Composition Spectrometer (NACS)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options: 'neutral1s'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars= load(instrument='nacs', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return nacs_postprocessing(tvars)


def nacs_postprocessing(variables):
    """
    Placeholder for NACS post-processing 
    """
    return variables


def rpa(trange=['1983-02-16', '1983-02-17'],
        datatype='ion2s',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Retarding Potential Analyzer (RPA)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options: 'ion2s'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars= load(instrument='rpa', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)


    if tvars is None or notplot or downloadonly:
        return tvars

    return rpa_postprocessing(tvars)


def rpa_postprocessing(variables):
    """
    Placeholder for RPA post-processing 
    """
    return variables


def fpi(trange=['1983-02-16', '1983-02-17'],
        datatype='8s',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Fabry-Pérot Interferometer (FPI)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                '8s' for 8-second resolution data

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='fpi', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return fpi_postprocessing(tvars)


def fpi_postprocessing(variables):
    """
    Placeholder for FPI post-processing 
    """
    return variables


def idm(trange=['1983-02-16', '1983-02-17'],
        datatype='250ms',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Ion Drift Meter (IDM)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options: '250ms'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='idm', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return idm_postprocessing(tvars)


def idm_postprocessing(variables):
    """
    Placeholder for IDM post-processing 
    """
    return variables


def wats(trange=['1983-02-16', '1983-02-17'],
        datatype='2s', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Wind and Temperature Spectrometer (WATS)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                '2s' for 2 second data

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='wats', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return wats_postprocessing(tvars)


def wats_postprocessing(variables):
    """
    Placeholder for WATS post-processing 
    """
    return variables


def vefi(trange=['1983-02-16', '1983-02-17'],
        datatype='ac500ms',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Vector Electric Field Instrument (VEFI)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'ac500ms', 'dca500ms'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='vefi', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return vefi_postprocessing(tvars)


def vefi_postprocessing(variables):
    """
    Placeholder for VEFI post-processing 
    """
    return variables


def lang(trange=['1983-02-16', '1983-02-17'],
        datatype='500ms',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Langmuir Probe Instrument (LANG)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options: '500ms'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='lang', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return lang_postprocessing(tvars)


def lang_postprocessing(variables):
    """
    Placeholder for LANG post-processing 
    """
    return variables


def datasets(instrument=None, label=True):
    return find_datasets(mission='Dynamics Explorer', instrument=instrument, label=label)



import csv
import os
import numpy as np
from pyspedas import time_double, time_string


def mms_read_feeps_sector_masks_csv(trange):
    """
    This function returns the FEEPS sectors to mask due to sunlight contamination
    
    Parameters
    -----------
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            
    Returns
    -----------
        Hash table containing the sectors to mask for each spacecraft and sensor ID

    """
    masks = {}

    dates = [1447200000.0000000, # 11/11/2015
             1468022400.0000000, # 7/9/2016
             1477612800.0000000, # 10/28/2016
             1496188800.0000000, # 5/31/2017
             1506988800.0000000, # 10/3/2017
             1538697600.0000000, # 10/5/2018
             1642032000.0000000, # 1/13/2022
             1651795200.0000000, # 5/6/2022
             1660521600.0000000] # 8/15/2022

    # find the file closest to the start time
    nearest_date = dates[(np.abs(np.array(dates)-time_double(trange[0]))).argmin()]

    for mms_sc in [1, 2, 3, 4]:
        csv_file = os.sep.join([os.path.dirname(os.path.abspath(__file__)), 'sun', 'MMS'+str(mms_sc)+'_FEEPS_ContaminatedSectors_'+time_string(nearest_date, fmt='%Y%m%d')+'.csv'])

        csv_file = open(csv_file, 'r')
        csv_reader = csv.reader(csv_file)
        csv_data = []

        for line in csv_reader:
            csv_data.append([float(l) for l in line])
            
        csv_file.close()

        csv_data = np.array(csv_data)

        for i in range(0, 12):
            mask_vals = []
            for val_idx in range(0, len(csv_data[:, i])):
                if csv_data[val_idx, i] == 1: mask_vals.append(val_idx)
            masks['mms'+str(mms_sc)+'imaskt'+str(i+1)] = mask_vals

        for i in range(0, 12):
            mask_vals = []
            for val_idx in range(0, len(csv_data[:, i+12])):
                if csv_data[val_idx, i+12] == 1: mask_vals.append(val_idx)
            masks['mms'+str(mms_sc)+'imaskb'+str(i+1)] = mask_vals

    return masks



import logging
from .spinmodel import Spinmodel, save_spinmodel
from pytplot import data_exists


def spinmodel_postprocess(probe: str):
    """ Create and initialize three Spinmodel objects using tplot variables loaded from the STATE CDFs.

        The three models correspond to the three available correction levels: 0 = no corrections, 1 = waveform
        corrections, 2 = spin fit corrections.

        Each of the three models is stored in a dictionary via the save_spinmodel routine.

    Args:
        probe (str):  A single letter string specifying the probe for the models being built.

    """

    # It is possible that /get_support_data was specified, but only a limited set of variables
    # was requested.  Check that all needed spin model variables are present before attempting
    # to create the models.

    sm_quantities = ['tend', 'spinper', 'c', 'phaserr', 'nspins', 'npts', 'maxgap', 'initial_delta_phi', 'idpu_spinper',
                     'segflags']
    missing_var = False
    for v in sm_quantities:
        non_ecl_v = 'th' + probe + '_spin_' + v
        ecl_v = 'th' + probe + '_spin_ecl_' + v
        if not (data_exists(non_ecl_v) and data_exists(ecl_v)):
            missing_var = True

    if missing_var:
        logging.warning("Some required spin model variables were not requested, skipping spin model creation")
        return

    logging.info("Creating spin model for probe " + probe + " correction level 0")
    # Expect a warning message here: That name is currently not in tplot.
    # Maybe there's a better idiom in pytplot for checking whether a variable exists?
    model0 = Spinmodel(probe, 0)
    save_spinmodel(probe, 0, model0)
    logging.info("Creating spin model for probe " + probe + " correction level 1")
    model1 = Spinmodel(probe, 1)
    save_spinmodel(probe, 1, model1)
    logging.info("Creating spin model for probe " + probe + " correction level 2")
    model2 = Spinmodel(probe, 2)
    save_spinmodel(probe, 2, model2)



import numpy as np

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum


def mms_pgs_make_phi_spec(data_in, resolution=32):
    """
    Builds phi (longitudinal) spectrogram from a sanitized particle data structure.

    Parameters
    ----------
    data_in : dict
        The sanitized particle data structure containing 'phi', 'data', and 'bins' arrays.
    resolution : int, optional
        The number of bins to divide the 360 degrees of phi into. Default is 32.

    Returns
    -------
    y : array
        The bin centers for the phi spectrogram.
    ave : array
        The phi spectrogram with shape (n_phi,).

    Notes
    -----
    This function concatenates the sample's data to the `spec` variable. Both
    the spectrogram `spec` and the y-axis `yaxis` will be initialized if not set.
    The y-axis will remain a single dimension until a change is detected in the data,
    at which point it will be expanded to two dimensions.
    """
    data = data_in.copy()
    n_phi = resolution

    # zero inactive bins to ensure areas with no data are represented as NaN
    zero_bins = np.argwhere(data['bins'] == 0)
    if zero_bins.size != 0:
        for item in zero_bins:
            data['data'][item[0], item[1]] = 0.0

    ave = np.zeros(n_phi)
    bin_size = 360.0/n_phi
    outbins = np.arange(0, 361, bin_size)

    phi_flat = data['phi'].flatten()
    data_flat = data['data'].flatten()
    bins_flat = data['bins'].flatten()

    for bin_idx in range(0, len(outbins)-1):
        this_bin = np.argwhere((phi_flat >= outbins[bin_idx]) & (phi_flat < outbins[bin_idx+1]))
        if len(this_bin) > 0:
            bins = nansum(bins_flat[this_bin])
            if bins != 0.0:
                ave[bin_idx] += nansum(data_flat[this_bin])/bins
    
    y = outbins[0:n_phi]+0.5*(outbins[1::]-outbins[0:n_phi])

    return y, ave



import logging
import pytplot


def tdotp(variable1, variable2, newname=None):
    """
        Routine to calculate the dot product of two tplot variables 
        containing arrays of vectors and storing the results in a 
        tplot variable

    Input
    -------
        variable1: str
            First tplot variable

        variable2: str
            Second tplot variable

    Parameters
    -----------
        newname: str
            Name of the output variable
            
    Returns
    --------
        Name of the tplot variable
    """
    logging.info("tdotp has been moved to the pytplot.tplot_math module. Please update your imports!")
    logging.info("This version will eventually be removed.")
    pytplot.tplot_math.tdotp(variable1=variable1,variable2=variable2,newname=newname)


"""
Creates a deep copy of a pytplot variable, with a new name.

Notes
-----
Allowed wildcards are ? for a single character, * from multiple characters.

"""
import logging
import pytplot
import pyspedas
import copy


def tcopy_one(name_in, name_out):
    """
    Copy a single pytplot variable.

    Parameters
    ----------
    name_in : str
        Pytplot name in.
    name_out : str
        Pytplot name out.

    Returns
    -------
    None.

    """
    # Copies one pytplot variable
    tvar_old = pytplot.data_quants[name_in]
    tvar_new = copy.deepcopy(tvar_old)
    tvar_new.name = name_out
    pytplot.data_quants.update({name_out: tvar_new})
    logging.info(name_in + ' copied to ' + name_out)


def tcopy(names_in, names_out=None, suffix=None):
    """
    Copy a list of pytplot variables.

    Parameters
    ----------
    names_in : str/list of str
        List of pytplot names.
    names_out : str/list of str, optional
        List of pytplot names. The default is None.
    suffix : str, optional
        Suffix to apply to names_in. The default is '-copy'.

    Returns
    -------
    None.

    """
    names_in = pyspedas.tnames(names_in)
    if len(names_in) < 1:
        logging.error('tcopy error: No pytplot variables found.')
        return

    if suffix is None:
        suffix = '-copy'

    if names_out is None:
        names_out = [s + suffix for s in names_in]

    if isinstance(names_out, str):
        names_out = [names_out]

    if len(names_in) != len(names_out):
        logging.error('tcopy error: List with the names_in does not match list\
              with the names out.')
        return

    for i in range(len(names_in)):
        n = names_in[i]
        o = names_out[i]
        if len(pyspedas.tnames(n)) == 1:
            tcopy_one(n, o)
        else:
            logging.error('tplot name not found: ' + n)




import numpy as np
from pyspedas.utilities.interpol import interpol

def gsm2lmn(times, Rxyz, Bxyz, swdata=None):
    '''
    Transforms vector field from GSM to LMN (boundary-normal) coordinate
    system for the magnetopause using the Shue et al. (1998) magnetopause model


    '''

    if swdata is None:
        # swdata is an array containing: [SW times, SW dynamic pressure, SW B-field]
        swdata = np.zeros((len(times), 3))
        swdata[:, 0] = times
        swdata[:, 1] = 2.088 # nPa
    else:
        timessw = swdata[:, 0]
        dparr = swdata[:, 1]
        bzarr = swdata[:, 2]
        # replace NaNs with a standard value
        for nanval in np.argwhere(np.isnan(dparr)):
            dparr[nanval[0]] = 2.088 # nPa
        for nanval in np.argwhere(np.isnan(bzarr)):
            bzarr[nanval[0]] = 0 # nT

        # interpolate (linear) the data from the SW times to the B-field times
        dparra = interpol(dparr, timessw, times)
        bzarra = interpol(bzarr, timessw, times)
        swdata = np.zeros((len(times), 3))
        swdata[:, 0] = times
        swdata[:, 1] = dparra
        swdata[:, 2] = bzarra

    Blmn = np.zeros((len(times), len(Bxyz[0, :])))

    # start the transformation
    # mostly heisted from the IDL version
    for i in np.arange(len(times)):
        bz = swdata[i, 2]
        dp = max(swdata[i, 1], 0.01)
        x = Rxyz[i, 0]
        y = Rxyz[i, 1]
        z = Rxyz[i, 2]

        alpha = (0.58 - 0.007*bz)*(1 + 0.024*np.log(dp))
        theta = np.arccos(x/np.sqrt(x**2+y**2+z**2))
        rho = np.sqrt(y**2+z**2)

        if rho > 0:
            tang1 = np.array([0.0, z, -y])
            tang2 = np.array([value*alpha*np.sin(theta)/(1+np.cos(theta)) for value in [x, y, z]]) + np.array([-rho**2/rho, x*y/rho, x*z/rho])
            tang1 = [value/np.sqrt(np.sum(np.square(tang1))) for value in tang1]
            tang2 = [value/np.sqrt(np.sum(np.square(tang2))) for value in tang2]

            dN = np.cross(tang1, tang2)
            dM = np.cross(dN, [0, 0, 1])
            dM = dM/np.sqrt(np.sum(np.square(dM)))
            dL = np.cross(dM, dN)
        else:
            dN = [1., 0., 0.]
            dM = [0., 1., 0.]
            dL = [0., 0., 1.]

        transm = np.array([dL, dM, dN])
        Blmn[i, :] = np.matmul(transm, Bxyz[i, :])

    return Blmn


import logging
import unittest
from pyspedas.themis import state
from pytplot import data_exists, get_data, del_data, tplot_restore
from numpy.testing import assert_allclose

class StateDataValidation(unittest.TestCase):
    """ Tests creation of support variables in themis.state() """

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The IDL script that creates data file: projects/themis/state/thm_state_validate.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.themis.config import CONFIG

        # Testing time range
        cls.t = ['2008-03-23', '2008-03-28']


        # Download validation file
        remote_server = 'https://spedas.org/'
        remote_name = 'testfiles/thm_state_validate.tplot'
        datafile = download(remote_file=remote_name,
                           remote_path=remote_server,
                           local_path=CONFIG['local_data_dir'],
                           no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        tplot_restore(filename)
        #pytplot.tplot_names()
        cls.tha_pos = get_data('tha_state_pos')
        cls.tha_vel = get_data('tha_state_vel')
        cls.tha_spinras = get_data('tha_state_spinras')
        cls.tha_spindec = get_data('tha_state_spindec')
        cls.tha_spinras_correction = get_data('tha_state_spinras_correction')
        cls.tha_spindec_correction = get_data('tha_state_spindec_correction')
        cls.tha_spinras_corrected = get_data('tha_state_spinras_corrected')
        cls.tha_spindec_corrected = get_data('tha_state_spindec_corrected')

        # Load with pyspedas
        state(probe='a',trange=cls.t,get_support_data=True)

    def setUp(self):
        """ We need to clean tplot variables before each run"""
        # del_data('*')

    def test_state_spinras(self):
        """Validate state variables """
        my_data = get_data('tha_state_spinras')
        assert_allclose(my_data.y,self.tha_spinras.y,rtol=1.0e-06)

    def test_state_spindec(self):
        """Validate state variables """
        my_data = get_data('tha_state_spindec')
        assert_allclose(my_data.y,self.tha_spindec.y,rtol=1.0e-06)

    def test_state_spinras_correction(self):
        """Validate state variables """
        my_data = get_data('tha_state_spinras_correction')
        assert_allclose(my_data.y,self.tha_spinras_correction.y,rtol=1.0e-06)

    def test_state_spindec_correction(self):
        """Validate state variables """
        my_data = get_data('tha_state_spindec_correction')
        assert_allclose(my_data.y,self.tha_spindec_correction.y,rtol=1.0e-06)

    def test_state_spinras_corrected(self):
        """Validate state variables """
        my_data = get_data('tha_state_spinras_corrected')
        assert_allclose(my_data.y,self.tha_spinras_corrected.y,rtol=1.0e-06)

    def test_state_spindec_corrected(self):
        """Validate state variables """
        my_data = get_data('tha_state_spindec_corrected')
        assert_allclose(my_data.y,self.tha_spindec_corrected.y,rtol=1.0e-06)

    def test_state_reload_no_v03(self):
        # Test overwriting of spin axis correction variables if data is loaded with V03 corrections, then other
        # data loaded that doesn't have the corrections (prevents dangling correction variables)
        ts1 = ['2007-03-23','2007-03-24']
        ts2 = ['2023-01-01','2023-01-02']
        state(trange=ts1,probe='a',get_support_data=True) # V03 corrections exist
        self.assertTrue(data_exists('tha_spinras_correction'))
        self.assertTrue(data_exists('tha_spindec_correction'))
        state(trange=ts2,probe='a',get_support_data=True) # V03 corrections do not exist
        self.assertFalse(data_exists('tha_spinras_correction'))
        self.assertFalse(data_exists('tha_spindec_correction'))

if __name__ == '__main__':
    unittest.main()



import logging
import numpy as np
from pytplot import get_data, store_data, options, tnames

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_eis_omni(probe, species='proton', datatype='extof', suffix='', data_units='flux', data_rate='srvy', level='l2'):
    """
    This function will calculate the omnidirectional EIS spectrograms, and is automatically called from mms_load_eis
    
    Parameters
    ----------
        probe: str
            probe #, e.g., '4' for MMS4

        species: str
            species for calculation (default: 'proton')

        datatype: str
            'extof' or 'phxtof' (default: 'extof')

        suffix: str
            suffix of the loaded data

        data_units: str
            'flux' or 'cps' (default: 'flux')

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst' (default: 'srvy')

        level: str
            data level ['l1a','l1b','l2pre','l2' (default)]

    Returns
    ---------
        Name of tplot variable created.
    """
    
    probe = str(probe)
    species_str = datatype + '_' + species
    prefix = 'mms' + probe + '_epd_eis_' + data_rate + '_' + level + '_'

    if data_units == 'flux':
        units_label = '1/(cm^2-sr-s-keV)'
    elif data_units == 'cps':
        units_label = '1/s'
    elif data_units == 'counts':
        units_label = 'counts'

    telescopes = tnames(pattern=prefix + species_str + '_*' + data_units + '_t?'+suffix)

    if len(telescopes) == 6:
        scope_data = get_data(telescopes[0])
            
        if len(scope_data) <= 2:
            logging.error("Error, couldn't find energy table for the variable: " + telescopes[0])
            return None

        time, data, energies = scope_data

        flux_omni = np.zeros((len(time), len(energies)))
        for t in telescopes:
            time, data, energies = get_data(t)
            flux_omni = flux_omni + data

        store_data(prefix + species_str + '_' + data_units + '_omni' + suffix, data={'x': time, 'y': flux_omni/6., 'v': energies})
        options(prefix + species_str + '_' + data_units + '_omni' + suffix, 'spec', 1)
        options(prefix + species_str + '_' + data_units + '_omni' + suffix, 'ylog', 1)
        options(prefix + species_str + '_' + data_units + '_omni' + suffix, 'zlog', 1)
        options(prefix + species_str + '_' + data_units + '_omni' + suffix, 'ztitle', units_label)
        options(prefix + species_str + '_' + data_units + '_omni' + suffix, 'ytitle', 'MMS' + probe + ' ' + datatype + ' ' + species)
        options(prefix + species_str + '_' + data_units + '_omni' + suffix, 'ysubtitle', 'Energy [keV]')
        options(prefix + species_str + '_' + data_units + '_omni' + suffix, 'yrange', [14, 45])

        # create new variable with omni energy limits
        energy_minus = get_data(prefix + species_str + '_t0_energy_dminus' + suffix)
        energy_gm = get_data(prefix + species_str + '_t0_energy' + suffix)
        energy_plus = get_data(prefix + species_str + '_t0_energy_dplus' + suffix)

        if isinstance(energy_minus, np.ndarray) and isinstance(energy_plus, np.ndarray):
            # transpose is used here to make the variable match the variable in IDL
            store_data(prefix + species_str + '_energy_range' + suffix, data={'y': np.array([energy_gm-energy_minus,energy_gm+energy_plus]).transpose()})

        return prefix + species_str + '_' + data_units + '_omni' + suffix
    else:
        logging.error('Error, problem finding the telescopes to calculate omni-directional spectrograms')
        return None



import os
import unittest
from pytplot import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_celias_data(self):
        out_vars = pyspedas.soho.celias(time_clip=True)
        self.assertTrue(data_exists('V_p'))
        self.assertTrue(data_exists('N_p'))

    def test_load_cosp_data(self):
        out_vars = pyspedas.soho.costep(time_clip=True)
        self.assertTrue(data_exists('P_int'))
        self.assertTrue(data_exists('He_int'))

    def test_load_erne_data(self):
        out_vars = pyspedas.soho.erne(time_clip=True)
        self.assertTrue(data_exists('PH'))

    def test_load_orbit_data(self):
        out_vars = pyspedas.soho.orbit(time_clip=True)
        self.assertTrue(data_exists('GSE_POS'))
        self.assertTrue(data_exists('GSE_VEL'))

    def test_load_notplot(self):
        out_vars = pyspedas.soho.erne(notplot=True)
        self.assertTrue('PH' in out_vars)

    def test_downloadonly(self):
        files = pyspedas.soho.erne(downloadonly=True, trange=['2006-06-01', '2006-06-02'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()

    


import logging
import numpy as np
from pyspedas import time_string, time_double, tnames
import pytplot


def mms_feeps_remove_bad_data(probe='1', data_rate='srvy', datatype='electron', level='l2', suffix='', trange=None):
    """
    This function removes bad eyes, bad lowest energy channels based on data from Drew Turner
    
    Parameters
    ------------
        probe: str
            probe #, e.g., '4' for MMS4

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        datatype: str
            'electron' or 'ion'

        level: str
            data level

        suffix: str
            suffix of the loaded data

        trange: list of str or list of float
            Time range
    Returns
    ------------
        None
    """
    if trange is None:
        logging.error('Time range required for mms_feeps_remove_bad_data.')
        return

    data_rate_level = data_rate + '_' + level

    # electrons first, remove bad eyes
    #;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 1. BAD EYES ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    #  First, here is a list of the EYES that are bad, we need to make sure these 
    #  data are not usable (i.e., make all of the counts/rate/flux data from these eyes NAN). 
    #  These are for all modes, burst and survey:
    bad_data_table = {}
    bad_data_table['2017-10-01'] = {}
    bad_data_table['2017-10-01']['mms1'] = {'top': [1], 'bottom': [1, 11]}
    bad_data_table['2017-10-01']['mms2'] = {'top': [5, 7, 12], 'bottom': [7]}
    bad_data_table['2017-10-01']['mms3'] = {'top': [2, 12], 'bottom': [2, 5, 11]}
    bad_data_table['2017-10-01']['mms4'] = {'top': [1, 2, 7], 'bottom': [2, 4, 5, 10, 11]}

    bad_data_table['2018-10-01'] = {}
    bad_data_table['2018-10-01']['mms1'] = {'top': [1], 'bottom': [1, 11]}
    bad_data_table['2018-10-01']['mms2'] = {'top': [7, 12], 'bottom': [2, 12]}
    bad_data_table['2018-10-01']['mms3'] = {'top': [1, 2], 'bottom': [5, 11]}
    bad_data_table['2018-10-01']['mms4'] = {'top': [1, 7], 'bottom': [4, 11]}

    dates = np.asarray(time_double(list(bad_data_table.keys())))
    closest_table_tm = (np.abs(dates - time_double(trange[0]))).argmin()

    closest_table = time_string(dates[closest_table_tm], '%Y-%m-%d')
    bad_data = bad_data_table[closest_table]['mms'+probe]

    bad_vars = []

    # top electrons
    for bad_var in bad_data['top']:
        if bad_var in [6, 7, 8]:
            continue  # ion eyes
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_counts_sensorid_'+str(bad_var)+suffix))

    # bottom electrons
    for bad_var in bad_data['bottom']:
        if bad_var in [6, 7, 8]:
            continue  # ion eyes
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_counts_sensorid_'+str(bad_var)+suffix))

    # top ions
    for bad_var in bad_data['top']:
        if bad_var not in [6, 7, 8]:
            continue  # ion eyes
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_counts_sensorid_'+str(bad_var)+suffix))

    # bottom ions
    for bad_var in bad_data['bottom']:
        if bad_var not in [6, 7, 8]:
            continue  # ion eyes
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_counts_sensorid_'+str(bad_var)+suffix))

    for bad_var in bad_vars:
        if not bad_var:
            continue
        bad_var_data = pytplot.get_data(bad_var[0])

        if bad_var_data is not None:
            times, data, energies = bad_var_data
            data[:] = np.nan
            pytplot.store_data(bad_var[0], data={'x': times, 'y': data, 'v': energies})

    # ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 2. BAD LOWEST E-CHANNELS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    # ; Next, these eyes have bad first channels (i.e., lowest energy channel, E-channel 0 in IDL indexing).  
    # ; Again, these data (just the counts/rate/flux from the lowest energy channel ONLY!!!) 
    # ; should be hardwired to be NAN for all modes (burst and both types of survey).  
    # ; The eyes not listed here or above are ok though... so once we do this, we can actually start 
    # ; showing the data down to the lowest levels (~33 keV), meaning we'll have to adjust the hard-coded 
    # ; ylim settings in SPEDAS and the SITL software:

    # ; from Drew Turner, 5Oct18:
    # ;Bad Channels (0 and 1):
    # ;Update: All channels 0 (Ch0) on MMS-2, -3, and -4 electron eyes (1, 2, 3, 4, 5, 9, 10, 11, 12) should be NaN
    # ;Additionally, the second channels (Ch1) on the following should also be made NaN:
    # ;MMS-1: Top: Ch0 on Eyes 6, 7
    # ;Bot: Ch0 on Eyes 6, 7, 8
    # ;MMS-2: Top:
    # ;Bot: Ch0 on Eyes 6, 8
    # ;MMS-3: Top: Ch0 on Eye 8
    # ;Bot: Ch0 on Eyes 6, 7
    # ;MMS-4: Top: Ch1 on Eye 1; Ch0 on Eye 8
    # ;Bot: Ch0 on Eyes 6, 7, 8; Ch1 on Eye 9

    bad_vars = []
    bad_vars_both_chans = []
    bad_vars_3_chans = []

    # -- the following are the pre-May 2019 masks
    # bad_ch0 = {}
    # bad_ch0['mms1'] = {'top': [2, 5, 6, 7], 'bottom': [2, 3, 4, 5, 6, 7, 8, 9, 11, 12]}
    # bad_ch0['mms2'] = {'top': [1, 2, 3, 4, 5, 9, 10, 11, 12], 'bottom': [1, 2, 3, 4, 5, 6, 8, 9, 10, 11, 12]}
    # bad_ch0['mms3'] = {'top': [1, 2, 3, 4, 5, 8, 9, 10, 11, 12], 'bottom': [1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 12]}
    # bad_ch0['mms4'] = {'top': [1, 2, 3, 4, 5, 8, 9, 10, 11, 12], 'bottom': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]}

    # bad_ch1 = {}
    # bad_ch1['mms1'] = {'top': [], 'bottom': [11]}
    # bad_ch1['mms2'] = {'top': [8], 'bottom': [12]}
    # bad_ch1['mms3'] = {'top': [1], 'bottom': []}
    # bad_ch1['mms4'] = {'top': [1], 'bottom': [6, 9]}

    # -- the following are the post-May 2019 masks
    bad_ch0 = {}
    bad_ch0['mms1'] = {'top': [2, 5, 6], 'bottom': [2, 3, 4, 5, 6, 8, 9, 11, 12]}
    bad_ch0['mms2'] = {'top': [1, 2, 3, 4, 5, 6, 8, 9, 10, 11, 12], 'bottom': [1, 2, 3, 4, 5, 7, 9, 10, 11, 12]}
    bad_ch0['mms3'] = {'top': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], 'bottom': [1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 12]}
    bad_ch0['mms4'] = {'top': [1, 2, 3, 4, 5, 6, 8, 9, 10, 11, 12], 'bottom': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]}

    bad_ch1 = {}
    bad_ch1['mms1'] = {'top': [6], 'bottom': [6, 11]}
    bad_ch1['mms2'] = {'top': [8], 'bottom': [7, 12]}
    bad_ch1['mms3'] = {'top': [1, 6, 7], 'bottom': [6, 7]}
    bad_ch1['mms4'] = {'top': [1, 6], 'bottom': [6, 7, 8, 9]}

    bad_ch2 = {}
    bad_ch2['mms1'] = {'top': [], 'bottom': []}
    bad_ch2['mms2'] = {'top': [], 'bottom': []}
    bad_ch2['mms3'] = {'top': [], 'bottom': []}
    bad_ch2['mms4'] = {'top': [], 'bottom': [6, 7]}

    bad_ch0 = bad_ch0['mms'+str(probe)]
    bad_ch1 = bad_ch1['mms'+str(probe)]
    bad_ch2 = bad_ch2['mms'+str(probe)]

    #### bottom channel

    # top electrons
    for bad_var in bad_ch0['top']:
        if bad_var in [6, 7, 8]:
            continue  # ion eyes
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_counts_sensorid_'+str(bad_var)+suffix))

    # bottom electrons
    for bad_var in bad_ch0['bottom']:
        if bad_var in [6, 7, 8]:
            continue  # ion eyes
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_counts_sensorid_'+str(bad_var)+suffix))

    # top ions
    for bad_var in bad_ch0['top']:
        if bad_var not in [6, 7, 8]:
            continue  # ion eyes
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_counts_sensorid_'+str(bad_var)+suffix))

    # bottom ions
    for bad_var in bad_ch0['bottom']:
        if bad_var not in [6, 7, 8]:
            continue  # ion eyes
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_counts_sensorid_'+str(bad_var)+suffix))

    #### bottom 2 channels

    # top electrons
    for bad_var in bad_ch1['top']:
        if bad_var in [6, 7, 8]:
            continue  # ion eyes
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_counts_sensorid_'+str(bad_var)+suffix))

    # bottom electrons
    for bad_var in bad_ch1['bottom']:
        if bad_var in [6, 7, 8]:
            continue  # ion eyes
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_counts_sensorid_'+str(bad_var)+suffix))

    # top ions
    for bad_var in bad_ch1['top']:
        if bad_var not in [6, 7, 8]:
            continue  # ion eyes
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_counts_sensorid_'+str(bad_var)+suffix))

    # bottom ions
    for bad_var in bad_ch1['bottom']:
        if bad_var not in [6, 7, 8]:
            continue  # ion eyes
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars_both_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_counts_sensorid_'+str(bad_var)+suffix))

    #### bottom 3 channels
    # top electrons
    for bad_var in bad_ch2['top']:
        if bad_var in [6, 7, 8]:
            continue  # ion eyes
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_top_counts_sensorid_'+str(bad_var)+suffix))

    # bottom electrons
    for bad_var in bad_ch2['bottom']:
        if bad_var in [6, 7, 8]:
            continue  # ion eyes
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_electron_bottom_counts_sensorid_'+str(bad_var)+suffix))

    # top ions
    for bad_var in bad_ch2['top']:
        if bad_var not in [6, 7, 8]:
            continue  # ion eyes
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_top_counts_sensorid_'+str(bad_var)+suffix))

    # bottom ions
    for bad_var in bad_ch2['bottom']:
        if bad_var not in [6, 7, 8]:
            continue  # ion eyes
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_count_rate_sensorid_'+str(bad_var)+suffix))
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_intensity_sensorid_'+str(bad_var)+suffix))
        bad_vars_3_chans.append(tnames('mms'+str(probe)+'_epd_feeps_'+data_rate_level+'_ion_bottom_counts_sensorid_'+str(bad_var)+suffix))

    # set the first energy channel to NaN
    for bad_var in bad_vars:
        if not bad_var:
            continue
        bad_var_data = pytplot.get_data(bad_var[0])
        if bad_var_data is not None:
            times, data, energies = bad_var_data

            # check if the energy table contains all nans
            if np.isnan(np.sum(energies)):
                continue

            data[:, 0] = np.nan
            pytplot.store_data(bad_var[0], data={'x': times, 'y': data, 'v': energies})

    # set the first and second energy channels to NaN
    for bad_var in bad_vars_both_chans:
        if not bad_var:
            continue
        bad_var_data = pytplot.get_data(bad_var[0])
        if bad_var_data is not None:
            times, data, energies = bad_var_data

            # check if the energy table contains all names
            if np.isnan(np.sum(energies)):
                continue

            data[:, 0] = np.nan
            data[:, 1] = np.nan
            pytplot.store_data(bad_var[0], data={'x': times, 'y': data, 'v': energies})

    # set the bottom 3 energy channels to NaN
    for bad_var in bad_vars_3_chans:
        if not bad_var:
            continue
        bad_var_data = pytplot.get_data(bad_var[0])
        if bad_var_data is not None:
            times, data, energies = bad_var_data

            # check if the energy table contains all names
            if np.isnan(np.sum(energies)):
                continue

            data[:, 0] = np.nan
            data[:, 1] = np.nan
            data[:, 2] = np.nan
            pytplot.store_data(bad_var[0], data={'x': times, 'y': data, 'v': energies})



import cdflib
import numpy as np

from pyspedas import tnames
from pytplot import get_data, store_data, options, clip, ylim, zlim

from ...satellite.erg.load import load


def isee_vlf(
    trange=['2017-03-30/12:00:00', '2017-03-30/15:00:00'],
    suffix='',
    site='all',
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    uname=None,
    passwd=None,
    time_clip=False,
    ror=True,
    cal_gain=False
):

    site_code_all = ['ath', 'gak', 'hus', 'ist', 'kap', 'mam', 'nai']

    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(' ')
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if 'all' in site_code:
        site_code = site_code_all
    
    site_code = list(set(site_code).intersection(site_code_all))

    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        prefix='isee_vlf_'+site_input+'_'
        file_res = 3600.
        pathformat = 'ground/vlf/'+site_input\
                        +'/%Y/%m/isee_vlf_'+site_input+'_%Y%m%d%H_v??.cdf'

        loaded_data_temp = load(pathformat=pathformat, file_res=file_res, trange=trange, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                        varformat=varformat, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)
        
        if notplot:
            loaded_data.update(loaded_data_temp)
        else:
            loaded_data += loaded_data_temp
        if (len(loaded_data_temp) > 0) and ror:
            try:
                if isinstance(loaded_data_temp, list):
                    if downloadonly:
                        cdf_file = cdflib.CDF(loaded_data_temp[0])
                        gatt = cdf_file.globalattsget()
                    else:
                        gatt = get_data(loaded_data_temp[0], metadata=True)['CDF']['GATT']
                elif isinstance(loaded_data_temp, dict):
                    gatt = loaded_data_temp[list(loaded_data_temp.keys())[0]]['CDF']['GATT']
                print('**************************************************************************')
                print(gatt["Logical_source_description"])
                print('')
                print(f'Information about {gatt["Station_code"]}')
                print(f'PI {gatt["PI_name"]}')
                print('')
                print(f'Affiliations: {gatt["PI_affiliation"]}')
                print('')
                print('Rules of the Road for ISEE VLF Data Use:')
                print('')
                for gatt_text in gatt["TEXT"]:
                    print(gatt_text)
                print({gatt["LINK_TEXT"]})
                print('**************************************************************************')
            except:
                print('printing PI info and rules of the road was failed')
            
        if (not downloadonly) and (not notplot):
            t_plot_name_list = list(set(tnames([prefix+'ch1'+suffix, 
                                    prefix+'ch2'+suffix])).intersection(loaded_data))
            options(t_plot_name_list, 'zlog', 1)
            options(t_plot_name_list, 'ytitle', 'Frequency [Hz]')
            options(t_plot_name_list, 'ysubtitle','')
            if not cal_gain:
                options(t_plot_name_list, 'ztitle', 'V^2/Hz')
            else:
                print('Calibrating the gain of VLF antenna system...')
                file_name = get_data(t_plot_name_list[0], metadata=True)['CDF']['FILENAME']
                if isinstance(file_name, list):
                    file_name = file_name[0]
                cdf_file = cdflib.CDF(file_name)
                
                ffreq=cdf_file.varget('freq_vlf')
                gain_ch1=cdf_file.varget('amplitude_cal_vlf_ch1')
                gain_ch2=cdf_file.varget('amplitude_cal_vlf_ch2')

                gain_ch1_mod = np.interp(ffreq, gain_ch1[0], gain_ch1[1]) * 1.e-9
                gain_ch2_mod = np.interp(ffreq, gain_ch2[0], gain_ch2[1]) * 1.e-9

                t_plot_name = prefix+'ch1' + suffix
                tmp1 = get_data(t_plot_name)
                if tmp1 is not None:
                    tmp1_metadata = get_data(t_plot_name, metadata=True)
                    tmp1_y = tmp1[1] / gain_ch1_mod/gain_ch1_mod
                    store_data(t_plot_name,
                                data={'x':tmp1[0],
                                      'y':tmp1_y,
                                      'v':tmp1[2]},
                                attr_dict=tmp1_metadata)

                t_plot_name = prefix+'ch2' + suffix
                tmp2 = get_data(t_plot_name)
                if tmp2 is not None:
                    tmp2_metadata = get_data(t_plot_name, metadata=True)
                    tmp2_y = tmp1[1] / gain_ch2_mod/gain_ch2_mod
                    store_data(t_plot_name,
                                data={'x':tmp2[0],
                                      'y':tmp2_y,
                                      'v':tmp2[2]},
                                attr_dict=tmp2_metadata)

                options(t_plot_name_list, 'ztitle', 'nT^2/Hz')



    return loaded_data



import pytplot
import logging

def data_exists(tvar):
    """
    Checks if a tplot variable exists
    """
    logging.info("data_exists has been moved to the pytplot package.  Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.data_exists(tvar=tvar)



from .load import load


def mag(trange=['2014-10-18', '2014-10-19'],
        datatype='sunstate-1sec', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Magnetometer (MAG)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'sunstate-1sec' for L2 data

        level: str
            Data level; options: 'l1', 'l2' (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='mag', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return mag_postprocessing(tvars)


def mag_postprocessing(variables):
    """
    Placeholder for MAG post-processing 
    """
    return variables


def swea(trange=['2014-10-18', '2014-10-19'],
        datatype='arc3d', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Solar Wind Electron Analyzer (SWEA)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'arc3d' for L2 data
                'arcpad' for L2 data
                'svy3d' for L2 data
                'svypad' for L2 data
                'svyspec' for L2 data

        level: str
            Data level; options: 'l2' (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='swea', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return swea_postprocessing(tvars)


def swea_postprocessing(variables):
    """
    Placeholder for SWEA post-processing 
    """
    return variables


def swia(trange=['2014-10-18', '2014-10-19'],
        datatype='onboardsvyspec', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Solar Wind Ion Analyzer (SWIA)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'onboardsvyspec' for L2 data
                'onboardsvymom' for L2 data
                'finesvy3d' for L2 data
                'finearc3d' for L2 data
                'coarsesvy3d' for L2 data
                'coarsearc3d' for L2 data

        level: str
            Data level; options: 'l2' (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='swia', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return swia_postprocessing(tvars)


def swia_postprocessing(variables):
    """
    Placeholder for SWIA post-processing 
    """
    return variables


def static(trange=['2014-10-18', '2014-10-19'],
        datatype='c0-64e2m', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the SupraThermal And Thermal Ion Composition (STATIC)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'c0-64e2m' for L2 data
                'c2-32e32m' for L2 data
                'c4-4e64m' for L2 data
                'c6-32e64m' for L2 data
                'c8-32e16d' for L2 data
                'ca-16e4d16a' for L2 data
                'cc-32e8d32m' for L2 data
                'cd-32e8d32m' for L2 data
                'ce-16e4d16a16m' for L2 data
                'cf-16e4d16a16m' for L2 data
                'd0-32e4d16a8m' for L2 data
                'd1-32e4d16a8m' for L2 data
                'd4-4d16a2m' for L2 data
                'd6-events' for L2 data
                'd7-fsthkp' for L2 data
                'd8-12r1e' for L2 data
                'd9-12r64e' for L2 data
                'da-1r64e' for L2 data
                'db-1024tof' for L2 data
                'hkp' for L2 data

        level: str
            Data level; options: 'l2' (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='static', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return static_postprocessing(tvars)


def static_postprocessing(variables):
    """
    Placeholder for STATIC post-processing 
    """
    return variables


def sep(trange=['2014-10-18', '2014-10-19'],
        datatype='s1-cal-svy-full', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Solar Energetic Particle (SEP) instrument
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                's1-cal-svy-full' for L2 data
                's2-cal-svy-full' for L2 data

        level: str
            Data level; options: 'l2' (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='sep', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return sep_postprocessing(tvars)


def sep_postprocessing(variables):
    """
    Placeholder for SEP post-processing 
    """
    return variables


def kp(trange=['2014-10-18', '2014-10-19'],
        datatype='kp-4sec', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Key Parameters (KP)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'kp-4sec' for 4 second KP data

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='kp', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return kp_postprocessing(tvars)


def kp_postprocessing(variables):
    """
    Placeholder for KP post-processing 
    """
    return variables



import logging
import pytplot

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_feeps_split_integral_ch(units_type, species, probe, suffix='', data_rate='srvy', level='l2', sensor_eyes=None):
    """
    This function splits the last integral channel from the FEEPS spectra, 
    creating 2 new tplot variables:

       [original variable]_clean - spectra with the integral channel removed
       [original variable]_500keV_int - the integral channel that was removed
    
    Parameters
    -----------
        units_type: str
            instrument datatype, e.g., 'intensity'

        species: str
            'electron' or 'ion'

        probe: str
            probe #, e.g., '4' for MMS4

        suffix: str
            suffix of the loaded data

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        level: str
            data level

        sensor_eyes: dict
            Hash table containing the active sensor eyes

    Returns
    -----------
        List of tplot variables created.
    """
    if sensor_eyes is None:
        logging.error('Error: sensor_eyes not defined')
        return

    out_vars = []

    top_sensors = sensor_eyes['top']
    bot_sensors = sensor_eyes['bottom']

    for sensor in top_sensors:
        top_name = 'mms'+str(probe)+'_epd_feeps_'+data_rate+'_'+level+'_'+species+'_top_'+units_type+'_sensorid_'+str(sensor)

        time, data, energies = pytplot.get_data(top_name+suffix)

        top_name_out = top_name+'_clean'+suffix
        try:
            pytplot.store_data(top_name_out, data={'x': time, 'y': data[:, :-1], 'v': energies[:-1]})
            pytplot.store_data(top_name+'_500keV_int'+suffix, data={'x': time, 'y': data[:, -1]})
            out_vars.append(top_name_out)
            out_vars.append(top_name+'_500keV_int'+suffix)
        except Warning:
            continue

        pytplot.del_data(top_name+suffix)

    if level == 'sitl': # SITL only has top sensors
        return

    for sensor in bot_sensors:
        bot_name = 'mms'+str(probe)+'_epd_feeps_'+data_rate+'_'+level+'_'+species+'_bottom_'+units_type+'_sensorid_'+str(sensor)

        time, data, energies = pytplot.get_data(bot_name+suffix)

        bot_name_out = bot_name+'_clean'+suffix
        try:
            pytplot.store_data(bot_name_out, data={'x': time, 'y': data[:, :-1], 'v': energies[:-1]})
            pytplot.store_data(bot_name+'_500keV_int'+suffix, data={'x': time, 'y': data[:, -1]})
            out_vars.append(bot_name_out)
            out_vars.append(bot_name+'_500keV_int'+suffix)
        except Warning:
            continue

        pytplot.del_data(bot_name+suffix)

    return out_vars



from pyspedas import tnames
from pytplot import options


def mms_eis_set_metadata(tplotnames, data_rate='srvy', datatype='extof', suffix=''):
    """
    This function updates the metadata for the EIS data products
    
    Parameters
    ----------
        tplotnames : list of str
            list of tplot variables loaded by the load routine

        data_rate : str 
            Data rate

        datatype : str 
            EIS datatype loaded (extof or phxtof)
            
        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

    """
    if datatype == 'extof':
        options(tnames('*_extof_proton_flux_omni*'), 'yrange', [55, 1000])
        # options(tnames('*_extof_alpha_flux_omni*'), 'yrange', [80, 650]) # removed in the latest files as of 3 Aug 2021
        options(tnames('*_extof_helium_flux_omni*'), 'yrange', [80, 650])
        options(tnames('*_extof_oxygen_flux_omni*'), 'yrange', [145, 950])
        options(tnames('*_extof_proton_flux_omni*'), 'x_interp', True)
        options(tnames('*_extof_proton_flux_omni*'), 'y_interp', True)
        options(tnames('*_extof_helium_flux_omni*'), 'x_interp', True)
        options(tnames('*_extof_helium_flux_omni*'), 'y_interp', True)
        options(tnames('*_extof_oxygen_flux_omni*'), 'x_interp', True)
        options(tnames('*_extof_oxygen_flux_omni*'), 'y_interp', True)



import unittest
from pyspedas.mms.particles.mms_part_getspec import mms_part_getspec
from pytplot import data_exists


class PGSTests(unittest.TestCase):
    def test_fpi_brst_fac_type(self):
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'],
                         data_rate='brst',
                         species='i',
                         fac_type='phigeo',
                         output='pa gyro moments')
        self.assertTrue(data_exists('mms1_dis_dist_brst_density'))
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'],
                         data_rate='brst',
                         species='i',
                         fac_type='xgse',
                         output='pa gyro moments')
        self.assertTrue(data_exists('mms1_dis_dist_brst_density'))

    def test_fpi_brst_i(self):
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'],
                         data_rate='brst',
                         species='i',
                         output='energy theta phi pa gyro moments')
        self.assertTrue(data_exists('mms1_dis_dist_brst_density'))
        self.assertTrue(data_exists('mms1_dis_dist_brst_velocity'))
        self.assertTrue(data_exists('mms1_dis_dist_brst_avgtemp'))
        self.assertTrue(data_exists('mms1_dis_dist_brst_energy'))
        self.assertTrue(data_exists('mms1_dis_dist_brst_theta'))
        self.assertTrue(data_exists('mms1_dis_dist_brst_phi'))
        self.assertTrue(data_exists('mms1_dis_dist_brst_pa'))
        self.assertTrue(data_exists('mms1_dis_dist_brst_gyro'))

    def test_fpi_disable_pe_corr(self):
        mms_part_getspec(trange=['2015-10-16/13:06:07', '2015-10-16/13:06:08'],
                         data_rate='brst',
                         species='e',
                         output='energy',
                         disable_photoelectron_corrections=True)
        self.assertTrue(data_exists('mms1_des_dist_brst_energy'))

    def test_fpi_brst_e(self):
        mms_part_getspec(trange=['2015-10-16/13:06:07', '2015-10-16/13:06:08'],
                         data_rate='brst',
                         species='e',
                         output='energy theta phi pa gyro moments')
        self.assertTrue(data_exists('mms1_des_dist_brst_density'))
        self.assertTrue(data_exists('mms1_des_dist_brst_velocity'))
        self.assertTrue(data_exists('mms1_des_dist_brst_avgtemp'))
        self.assertTrue(data_exists('mms1_des_dist_brst_energy'))
        self.assertTrue(data_exists('mms1_des_dist_brst_theta'))
        self.assertTrue(data_exists('mms1_des_dist_brst_phi'))
        self.assertTrue(data_exists('mms1_des_dist_brst_pa'))
        self.assertTrue(data_exists('mms1_des_dist_brst_gyro'))

    def test_fpi_brst_limits(self):
        mms_part_getspec(trange=['2015-10-16/13:06:07', '2015-10-16/13:06:08'],
                         data_rate='brst',
                         species='e',
                         theta=[0, 90],
                         phi=[0, 100],
                         gyro=[0, 180],
                         pitch=[45, 75],
                         energy=[1000, 20000],
                         output='energy theta phi pa gyro moments')
        self.assertTrue(data_exists('mms1_des_dist_brst_density'))
        self.assertTrue(data_exists('mms1_des_dist_brst_velocity'))
        self.assertTrue(data_exists('mms1_des_dist_brst_avgtemp'))
        self.assertTrue(data_exists('mms1_des_dist_brst_energy'))
        self.assertTrue(data_exists('mms1_des_dist_brst_theta'))
        self.assertTrue(data_exists('mms1_des_dist_brst_phi'))
        self.assertTrue(data_exists('mms1_des_dist_brst_pa'))
        self.assertTrue(data_exists('mms1_des_dist_brst_gyro'))

    def test_hpca_srvy_hplus(self):
        mms_part_getspec(trange=['2015-10-16/13:05', '2015-10-16/13:10'],
                         #data_rate='brst',
                         species='hplus',
                         instrument='hpca',
                         output='energy theta phi pa gyro moments')
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_density'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_velocity'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_avgtemp'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_energy'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_theta'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_phi'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_pa'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_gyro'))

    def test_hpca_srvy_oplus(self):
        mms_part_getspec(trange=['2015-10-16/13:05', '2015-10-16/13:10'],
                         species='oplus',
                         instrument='hpca',
                         output='energy')
        self.assertTrue(data_exists('mms1_hpca_oplus_phase_space_density_energy'))

    def test_hpca_srvy_heplus(self):
        mms_part_getspec(trange=['2015-10-16/13:05', '2015-10-16/13:10'],
                         species='heplus',
                         instrument='hpca',
                         output='energy')
        self.assertTrue(data_exists('mms1_hpca_heplus_phase_space_density_energy'))

    def test_hpca_srvy_heplusplus(self):
        mms_part_getspec(trange=['2015-10-16/13:05', '2015-10-16/13:10'],
                         species='heplusplus',
                         instrument='hpca',
                         output='energy')
        self.assertTrue(data_exists('mms1_hpca_heplusplus_phase_space_density_energy'))


if __name__ == '__main__':
    unittest.main()



import warnings
import astropy

from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG

def load(trange=['2013-11-5', '2013-11-6'], 
         instrument='fgm',
         datatype='k0', 
         suffix='', 
         get_support_data=False, 
         varformat=None,
         downloadonly=False,
         notplot=False,
         no_update=False,
         varnames=[],
         time_clip=False):
    """
    This function loads data from the ACE mission; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.ace.mfi
        pyspedas.ace.swe
        pyspedas.ace.epam
        pyspedas.ace.cris
        pyspedas.ace.sis
        pyspedas.ace.uleis
        pyspedas.ace.sepica
        pyspedas.ace.swics

    """

    if instrument == 'fgm':
        pathformat = 'mag/level_2_cdaweb/mfi_'+datatype+'/%Y/ac_'+datatype+'_mfi_%Y%m%d_v??.cdf'
    elif instrument == 'swe':
        pathformat = 'swepam/level_2_cdaweb/swe_'+datatype+'/%Y/ac_'+datatype+'_swe_%Y%m%d_v??.cdf'
    elif instrument == 'epm':
        pathformat = 'epam/level_2_cdaweb/epm_'+datatype+'/%Y/ac_'+datatype+'_epm_%Y%m%d_v??.cdf'
    elif instrument == 'cris':
        pathformat = 'cris/level_2_cdaweb/cris_'+datatype+'/%Y/ac_'+datatype+'_cris_%Y%m%d_v??.cdf'
    elif instrument == 'sis':
        pathformat = 'sis/level_2_cdaweb/sis_'+datatype+'/%Y/ac_'+datatype+'_sis_%Y%m%d_v??.cdf'
    elif instrument == 'ule':
        pathformat = 'uleis/level_2_cdaweb/ule_'+datatype+'/%Y/ac_'+datatype+'_ule_%Y%m%d_v??.cdf'
    elif instrument == 'sep':
        pathformat = 'sepica/level_2_cdaweb/sep_'+datatype+'/%Y/ac_'+datatype+'_sep_%Y%m%d_v??.cdf'
    elif instrument == 'swics':
        filename_dtype = datatype.split('_')[1] + '_' + datatype.split('_')[0]
        pathformat = 'swics/level_2_cdaweb/'+datatype+'/%Y/ac_'+filename_dtype+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    with warnings.catch_warnings():
        # for some reason, ACE CDFs throw ERFA warnings (likely while converting
        # times inside astropy); we're ignoring these here
        # see: https://github.com/astropy/astropy/issues/9603
        warnings.simplefilter('ignore', astropy.utils.exceptions.ErfaWarning)
        tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import cdflib
import numpy as np

from copy import deepcopy
from pytplot import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load


def gmag_isee_induction(
    trange=['2018-10-18/00:00:00','2018-10-18/02:00:00'],
    suffix='',
    site='all',
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    uname=None,
    passwd=None,
    time_clip=False,
    ror=True,
    frequency_dependent=False
):

    site_code_all = ['ath', 'gak', 'hus', 'kap', 'lcl', 'mgd', 'msr', 'nai', 'ptk', 'rik', 'sta', 'zgn']



    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(' ')
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if 'all' in site_code:
        site_code = site_code_all

    site_code = list(set(site_code).intersection(site_code_all))

    if frequency_dependent:
        frequency_dependent_structure = {}
        for site_input in site_code:
            frequency_dependent_structure[site_input] = {
                'site_code':'',
                'nfreq':0,
                'frequency':np.zeros(shape=(64)),
                'sensitivity':np.zeros(shape=(64,3)),
                'phase_difference':np.zeros(shape=(64,3))
            }

    prefix = 'isee_induction_'
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    
    for site_input in site_code:

        file_res = 3600.
        pathformat = 'ground/geomag/isee/induction/'+site_input\
                        +'/%Y/%m/isee_induction_'+site_input+'_%Y%m%d%H_v??.cdf'

        loaded_data_temp = load(pathformat=pathformat, file_res=file_res, trange=trange, prefix=prefix, suffix='_'+site_input+suffix, get_support_data=get_support_data,
                        varformat=varformat, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)
        
        if notplot:
            loaded_data.update(loaded_data_temp)
        else:
            loaded_data += loaded_data_temp
        if (len(loaded_data_temp) > 0) and ror:
            try:
                if isinstance(loaded_data_temp, list):
                    if downloadonly:
                        cdf_file = cdflib.CDF(loaded_data_temp[-1])
                        gatt = cdf_file.globalattsget()
                    else:
                        gatt = get_data(loaded_data_temp[-1], metadata=True)['CDF']['GATT']
                elif isinstance(loaded_data_temp, dict):
                    gatt = loaded_data_temp[list(loaded_data_temp.keys())[-1]]['CDF']['GATT']
                print('**************************************************************************')
                print(gatt["Logical_source_description"])
                print('')
                print(f'Information about {gatt["Station_code"]}')
                print('PI and Host PI(s):')
                print(gatt["PI_name"])
                print('')
                print('Affiliation: ')
                print(gatt["PI_affiliation"])
                print('')
                print('Rules of the Road for ISEE Induction Magnetometer Data Use:')
                for gatt_text in gatt["TEXT"]:
                    print(gatt_text)
                print(gatt["LINK_TEXT"])
                print('**************************************************************************')
            except:
                print('printing PI info and rules of the road was failed')
            
        
        if (not downloadonly) and (not notplot):

            tplot_name = prefix+'db_dt_' + site_input+suffix
            if tplot_name in loaded_data:
                clip(tplot_name, -1e+4, 1e+4)
                get_data_vars = get_data(tplot_name)
                ylim(tplot_name, np.nanmin(get_data_vars[1]), np.nanmax(get_data_vars[1]))
                options(tplot_name, 'legend_names', ['dH/dt','dD/dt','dZ/dt'])
                options(tplot_name, 'Color', ['b', 'g', 'r'])
                options(tplot_name, 'ytitle', '\n'.join(tplot_name.split('_')))
                
                if frequency_dependent:
                    
                    meta_data_var = get_data(tplot_name,metadata=True)

                    if meta_data_var is not None:
                        cdf_file = cdflib.CDF(meta_data_var['CDF']['FILENAME'])
                        cdfcont = cdf_file.varget('frequency', inq=True)
                        ffreq = cdf_file.varget('frequency')
                        ssensi=cdf_file.varget('sensitivity')
                        pphase=cdf_file.varget('phase_difference')
                        frequency_dependent_structure[site_input]['site_code'] = site_input
                        frequency_dependent_structure[site_input]['nfreq'] = cdfcont['max_records'] + 1
                        frequency_dependent_structure[site_input]['frequency'][0:ffreq.shape[0]] = deepcopy(ffreq)
                        frequency_dependent_structure[site_input]['sensitivity'][0:ssensi.shape[0]] = deepcopy(ssensi)
                        frequency_dependent_structure[site_input]['phase_difference'][0:pphase.shape[0]] = deepcopy(pphase)

    if frequency_dependent:
        return frequency_dependent_structure
    else:
        return loaded_data



from pyspedas.mms.mms_load_data import mms_load_data
from pyspedas.mms.mec.mms_mec_set_metadata import mms_mec_set_metadata
from pyspedas.mms.print_vars import print_vars
from pyspedas.mms.mms_config import CONFIG


@print_vars
def mms_load_mec(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='srvy', 
    level='l2', datatype='epht89q', varformat=None, varnames=[], suffix='', get_support_data=False,
    time_clip=False, no_update=False, available=False, notplot=False, latest_version=False, 
    major_version=False, min_version=None, cdf_version=None, spdf=False, always_prompt=False):
    """
    Load the attitude/ephemeris data from the LANL MEC files
    
    Parameters
    ----------
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4']. 

        data_rate : str or list of str
            instrument data rates for MEC include ['brst', 'srvy']. The
            default is 'srvy'.

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'

        datatype : str or list of str
            Valid datatypes for MEC are: ['ephts04d', 'epht89q', 'epht89d']; default is 'epht89q'

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.
            
        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multidimensional data products)

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested parameters

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')

        min_version: str
            Specify a minimum CDF version # to load

        latest_version: bool
            Only grab the latest CDF version in the requested time interval

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidentally save an incorrect password, or if your SDC password has changed

        spdf: bool
            If True, download the data from the SPDF instead of the SDC
    
    Notes
    ---------
        The default datatype was changed to 'epht89q' on 15Nov2021. There are sometimes issues with 
        creating the Tsyganenko 04 data products, which leads to the 'epht04d' files not being available. 
        The 'epht89d' files contain the same ephemeris data - the only difference are the data products
        that rely on the field model. 
        
    Returns
    -----------
        List of tplot variables created.

    """
    tvars = mms_load_data(trange=trange, probe=probe, data_rate=data_rate, level=level, instrument='mec',
            datatype=datatype, get_support_data=get_support_data, varformat=varformat, varnames=varnames, suffix=suffix,
            time_clip=time_clip, no_update=no_update, available=available, notplot=notplot, 
            latest_version=latest_version, major_version=major_version, min_version=min_version, 
            cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)

    if tvars is None or available or notplot or CONFIG['download_only']:
        return tvars

    mms_mec_set_metadata(probe, data_rate, level, suffix=suffix)

    return tvars



"""Tests of twavpol functions."""
import pytplot.get_data
from pytplot.importers.cdf_to_tplot import cdf_to_tplot
import unittest
import pytplot
from numpy.testing import assert_array_almost_equal_nulp, assert_array_max_ulp, assert_allclose
import numpy as np

from pytplot import data_exists

from pyspedas.analysis.twavpol import twavpol

class TwavpolDataValidation(unittest.TestCase):
    """ Compares cotrans results between Python and IDL """

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates the file: projects/themis/state/cotrans/thm_cotrans_validate.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.themis.config import CONFIG

        # Testing tolerance
        cls.tol = 1e-10

        # Download tplot files
        remote_server = 'https://spedas.org/'
        # remote_name = 'testfiles/thm_cotrans_validate.cdf'
        remote_name = 'testfiles/thc_twavpol_validate.tplot'
        datafile = download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path=CONFIG['local_data_dir'],
                            no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        pytplot.del_data('*')
        filename = datafile[0]
        # pytplot.cdf_to_tplot(filename)
        pytplot.tplot_restore(filename)
        pytplot.tplot_names()
        #pytplot.tplot('thc_scf_fac')
        #pytplot.tplot('thc_scf_fac_powspec')
        #pytplot.tplot('thc_scf_fac')
        cls.thc_scf_fac = pytplot.get_data('thc_scf_fac')
        cls.thc_scf_fac_attr = pytplot.get_data('thc_scf_fac',metadata=True)
        cls.thc_scf_fac_powspec = pytplot.get_data('thc_scf_fac_powspec')
        cls.thc_scf_fac_powspec_attr = pytplot.get_data('thc_scf_fac_powspec',metadata=True)
        cls.thc_scf_fac_degpol = pytplot.get_data('thc_scf_fac_degpol')
        cls.thc_scf_fac_waveangle = pytplot.get_data('thc_scf_fac_waveangle')
        cls.thc_scf_fac_elliptict = pytplot.get_data('thc_scf_fac_elliptict')
        cls.thc_scf_fac_helict = pytplot.get_data('thc_scf_fac_helict')

        twavpol('thc_scf_fac')


    def setUp(self):
        """ We need to clean tplot variables before each run"""
        # pytplot.del_data('*')

    def test_powspec(self):
        """ Validate twavpol power spectrum output """

        py_powspec = pytplot.get_data('thc_scf_fac_powspec')
        #print(np.nanmin(py_powspec.y),np.nanmax(py_powspec.y))
        assert_allclose(py_powspec.times,self.thc_scf_fac_powspec.times,atol=1.0e-06)
        assert_allclose(py_powspec.y, self.thc_scf_fac_powspec.y, atol=1.0e-06,rtol=1.0e-06)
        #pytplot.tplot('thc_scf_fac_powspec')

    def test_degpol(self):
        """ Validate twavpol degpol output """

        py_degpol = pytplot.get_data('thc_scf_fac_degpol')
        #print(np.min(py_degpol.y),np.nanmax(py_degpol.y))
        assert_allclose(py_degpol.times,self.thc_scf_fac_degpol.times,atol=1.0e-06)
        assert_allclose(py_degpol.y, self.thc_scf_fac_degpol.y, atol=1.0e-06,rtol=1.0e-06)
        #pytplot.tplot('thc_scf_fac_degpol')

    def test_waveangle(self):
        """ Validate twavpol waveangle output """

        py_waveangle = pytplot.get_data('thc_scf_fac_waveangle')
        #print(np.nanmin(py_waveangle.y),np.nanmax(py_waveangle.y))
        assert_allclose(py_waveangle.times,self.thc_scf_fac_waveangle.times,atol=1.0e-05)
        assert_allclose(py_waveangle.y, self.thc_scf_fac_waveangle.y, atol=1.0e-05,rtol=1.0e-06)
        #pytplot.tplot('thc_scf_fac_waveangle')

    def test_elliptict(self):
        """ Validate twavpol elliptict output """

        py_elliptict = pytplot.get_data('thc_scf_fac_elliptict')
        #print(np.nanmin(py_elliptict.y),np.nanmax(py_elliptict.y))
        assert_allclose(py_elliptict.times,self.thc_scf_fac_elliptict.times,atol=1.0e-06)
        assert_allclose(py_elliptict.y, self.thc_scf_fac_elliptict.y, atol=1.0e-06,rtol=1.0e-06)
        #pytplot.tplot('thc_scf_fac_elliptict')

    def test_helict(self):
        """ Validate twavpol helict output """

        py_helict = pytplot.get_data('thc_scf_fac_helict')
        #print(np.nanmin(py_helict.y),np.nanmax(py_helict.y))
        assert_allclose(py_helict.times,self.thc_scf_fac_helict.times,atol=1.0e-06)
        assert_allclose(py_helict.y, self.thc_scf_fac_helict.y, atol=1.0e-06,rtol=1.0e-06)
        #pytplot.tplot('thc_scf_fac_helict')


if __name__ == '__main__':
    unittest.main()
